import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, map, path, pipe, always, assocPath, prop, identity } from './fn.js'
import { Relationship } from './query-parser/types.js'

const nodeComponent = component => node => {
  const label = node.label?.value
  const name = node.name?.value
  const type = label && component[node.label.value] || identity

  return { name, label, type }
}

const notEntity = ({ label }) => label != null && label !== 'Entity'

const prepareQuery = (query, component) => {
  const { match, returnValues } = query
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = map(q => prepareQuery({ match: q, returnValues }, component), [f, t, e])

    return { from, to, edge, relationship, schema: query }
  } else {
    const components = match.components.map(nodeComponent(component))
    const types = components.filter(notEntity).map(prop('type'))
    return { components, relationship, getReturns: returnValues.from(components), query: defineQuery(types), schema: query }
  }
}

const hasComponents = (world, types, entity) => (
  types.every(type => hasComponent(world, type, entity))
)

const matches = (world, types, results, set = new Set()) => entity => {
  if (set.has(entity)) {
    return true
  } else if (hasComponents(world, types, entity) && (results == null || results.includes(entity))) {
    set.add(entity)
    return true
  } else {
    return false
  }
}

const queryRelationship = (query, world) => {
  const [from, edge, to] = ['from', 'edge', 'to'].map(name => {
    const node = query[name]
    return node.relationship ? node.from : node
  })

  const matchesFrom = matches(world, from.components.map(prop('type')), from.query(world))
  const matchesTo = matches(world, to.components.map(prop('type')))

  const Edge = edge.components[0].type
  const edges = edge.query(world)
  const result = []
  for (let i = 0; i < edges.length; i++) {
    const eid = edges[i]

    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const assembleQuery = (query, entities) => map(entity => ([{ entity, query }]), entities)

const executeQuery = curry((query, world) => {
  if (!query.relationship) {
    return assembleQuery(query, query.query(world))
  }

  const edges = queryRelationship(query, world)

  if (query.to.relationship) {
    return edges.reduce((acc, eid) => {
      const next = assocPath(
        'from.query',
        always(of(query.edge.components[0].type.to[eid])),
        query.to
      )

      return executeQuery(next, world)
        .map(child => [eid, ...child])
        .map(([l, r]) => [{ entity: l, query }, r])
        .concat(acc)
    }, [])
  } else {
    return assembleQuery(query, edges)
  }
})


const resolveNode = curry((node, obj) => {
  const { entity, query: { getReturns } } = node
  return { ...obj, ...getReturns(entity) }
})

const resolveRelationship = edges => {
  const { entity, query } = edges
  const Edge = query.edge.components[0].type

  const to = query.to?.from ?? query.to
  return pipe(
    resolveNode({ entity: Edge.from[entity], query: query.from }),
    resolveNode({ entity, query: query.edge }),
    resolveNode({ entity: Edge.to[entity], query: to }),
  )({})
}

const resolveReturns = query => {
  return query.map(edges => {
    const returns = edges.map(edge => {
      if (edge.query.relationship) {
        return resolveRelationship(edge)
      } else {
        return resolveNode(edge, {})
      }
    })

    return Object.assign({}, ...returns)
  })
}

export const query = curry((input, { world, component }) => {
  return parseAll(q, input).map(({ use, ...rest }) => {
    const useWorld = world[use?.value ?? 'default']
    const preparedQuery = prepareQuery(rest, component)
    const results = executeQuery(preparedQuery, useWorld)
    const returns = resolveReturns(results)
    return returns
  })
})

